@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the tools applications of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qqmltypecompiler_p.h"
 
@@ -53,6 +17,8 @@
 
 QT_BEGIN_NAMESPACE
 
+Q_LOGGING_CATEGORY(lcQmlTypeCompiler, "qt.qml.typecompiler");
+
 QQmlTypeCompiler::QQmlTypeCompiler(QQmlEnginePrivate *engine, QQmlTypeData *typeData,
                                    QmlIR::Document *parsedQML, const QQmlRefPointer<QQmlTypeNameCache> &typeNameCache,
                                    QV4::ResolvedTypeReferenceMap *resolvedTypeCache, const QV4::CompiledData::DependentTypesHasher &dependencyHasher)
@@ -100,7 +66,7 @@ QQmlRefPointer<QV4::ExecutableCompilationUnit> QQmlTypeCompiler::compile()
                 QQmlComponentAndAliasResolver resolver(this);
                 if (!resolver.resolve(result.processedRoot))
                     return nullptr;
-                pendingGroupPropertyBindings.resolveMissingPropertyCaches(engine, &m_propertyCaches);
+                pendingGroupPropertyBindings.resolveMissingPropertyCaches(&m_propertyCaches);
                 pendingGroupPropertyBindings.clear(); // anything that can be processed is now processed
             }
         } while (result.canResume);
@@ -151,11 +117,13 @@ QQmlRefPointer<QV4::ExecutableCompilationUnit> QQmlTypeCompiler::compile()
         document->jsModule.fileName = typeData->urlString();
         document->jsModule.finalUrl = typeData->finalUrlString();
         QmlIR::JSCodeGen v4CodeGenerator(document, engine->v4engine()->illegalNames());
-        if (!v4CodeGenerator.generateCodeForComponents(componentRoots())) {
-            recordError(v4CodeGenerator.error());
-            return nullptr;
+        for (QmlIR::Object *object : qAsConst(document->objects)) {
+            if (!v4CodeGenerator.generateRuntimeFunctions(object)) {
+                Q_ASSERT(v4CodeGenerator.hasError());
+                recordError(v4CodeGenerator.error());
+                return nullptr;
+            }
         }
-
         document->javaScriptCompilationUnit = v4CodeGenerator.generateCompilationUnit(/*generated unit data*/false);
     }
 
@@ -226,7 +194,7 @@ const QV4::CompiledData::Unit *QQmlTypeCompiler::qmlUnit() const
 
 const QQmlImports *QQmlTypeCompiler::imports() const
 {
-    return &typeData->imports();
+    return typeData->imports();
 }
 
 QVector<QmlIR::Object *> *QQmlTypeCompiler::qmlObjects() const
@@ -316,7 +284,7 @@ bool SignalHandlerResolver::resolveSignalHandlerExpressions()
 {
     for (int objectIndex = 0; objectIndex < qmlObjects.count(); ++objectIndex) {
         const QmlIR::Object * const obj = qmlObjects.at(objectIndex);
-        QQmlPropertyCache *cache = propertyCaches->at(objectIndex);
+        QQmlPropertyCache::ConstPtr cache = propertyCaches->at(objectIndex);
         if (!cache)
             continue;
         if (QQmlCustomParser *customParser = customParsers.value(obj->inheritedTypeNameIndex)) {
@@ -330,9 +298,9 @@ bool SignalHandlerResolver::resolveSignalHandlerExpressions()
     return true;
 }
 
-bool SignalHandlerResolver::resolveSignalHandlerExpressions(const QmlIR::Object *obj,
-                                                            const QString &typeName,
-                                                            QQmlPropertyCache *propertyCache)
+bool SignalHandlerResolver::resolveSignalHandlerExpressions(
+        const QmlIR::Object *obj, const QString &typeName,
+        const QQmlPropertyCache::ConstPtr &propertyCache)
 {
     // map from signal name defined in qml itself to list of parameters
     QHash<QString, QStringList> customSignals;
@@ -351,8 +319,8 @@ bool SignalHandlerResolver::resolveSignalHandlerExpressions(const QmlIR::Object
             const QMetaObject *attachedType = type.attachedPropertiesType(enginePrivate);
             if (!attachedType)
                 COMPILE_EXCEPTION(binding, tr("Non-existent attached object"));
-            QQmlRefPointer<QQmlPropertyCache> cache = compiler->enginePrivate()->cache(attachedType);
-            if (!resolveSignalHandlerExpressions(attachedObj, bindingPropertyName, cache.data()))
+            QQmlPropertyCache::ConstPtr cache = QQmlMetaType::propertyCache(attachedType);
+            if (!resolveSignalHandlerExpressions(attachedObj, bindingPropertyName, cache))
                 return false;
             continue;
         }
@@ -370,9 +338,9 @@ bool SignalHandlerResolver::resolveSignalHandlerExpressions(const QmlIR::Object
             qPropertyName = signalName.mid(0, signalName.length() - static_cast<int>(strlen("Changed")));
 
         bool notInRevision = false;
-        QQmlPropertyData * const signal = resolver.signal(signalName, &notInRevision);
-        QQmlPropertyData * const signalPropertyData = resolver.property(signalName, /*notInRevision ptr*/nullptr);
-        QQmlPropertyData * const qPropertyData = !qPropertyName.isEmpty() ? resolver.property(qPropertyName) : nullptr;
+        const QQmlPropertyData * const signal = resolver.signal(signalName, &notInRevision);
+        const QQmlPropertyData * const signalPropertyData = resolver.property(signalName, /*notInRevision ptr*/nullptr);
+        const QQmlPropertyData * const qPropertyData = !qPropertyName.isEmpty() ? resolver.property(qPropertyName) : nullptr;
         QString finalSignalHandlerPropertyName = signalName;
         QV4::CompiledData::Binding::Flag flag
                 = QV4::CompiledData::Binding::IsSignalHandlerExpression;
@@ -475,7 +443,7 @@ QQmlEnumTypeResolver::QQmlEnumTypeResolver(QQmlTypeCompiler *typeCompiler)
 bool QQmlEnumTypeResolver::resolveEnumBindings()
 {
     for (int i = 0; i < qmlObjects.count(); ++i) {
-        QQmlPropertyCache *propertyCache = propertyCaches->at(i);
+        QQmlPropertyCache::ConstPtr propertyCache = propertyCaches->at(i);
         if (!propertyCache)
             continue;
         const QmlIR::Object *obj = qmlObjects.at(i);
@@ -494,8 +462,8 @@ bool QQmlEnumTypeResolver::resolveEnumBindings()
 
             const QString propertyName = stringAt(binding->propertyNameIndex);
             bool notInRevision = false;
-            QQmlPropertyData *pd = resolver.property(propertyName, &notInRevision);
-            if (!pd)
+            const QQmlPropertyData *pd = resolver.property(propertyName, &notInRevision);
+            if (!pd || pd->isQList())
                 continue;
 
             if (!pd->isEnum() && pd->propType().id() != QMetaType::Int)
@@ -518,7 +486,9 @@ bool QQmlEnumTypeResolver::assignEnumToBinding(QmlIR::Binding *binding, QStringV
     return true;
 }
 
-bool QQmlEnumTypeResolver::tryQualifiedEnumAssignment(const QmlIR::Object *obj, const QQmlPropertyCache *propertyCache, const QQmlPropertyData *prop, QmlIR::Binding *binding)
+bool QQmlEnumTypeResolver::tryQualifiedEnumAssignment(
+        const QmlIR::Object *obj, const QQmlPropertyCache::ConstPtr &propertyCache,
+        const QQmlPropertyData *prop, QmlIR::Binding *binding)
 {
     bool isIntProp = (prop->propType().id() == QMetaType::Int) && !prop->isEnum();
     if (!prop->isEnum() && !isIntProp)
@@ -705,20 +675,20 @@ QQmlAliasAnnotator::QQmlAliasAnnotator(QQmlTypeCompiler *typeCompiler)
 void QQmlAliasAnnotator::annotateBindingsToAliases()
 {
     for (int i = 0; i < qmlObjects.count(); ++i) {
-        QQmlPropertyCache *propertyCache = propertyCaches->at(i);
+        QQmlPropertyCache::ConstPtr propertyCache = propertyCaches->at(i);
         if (!propertyCache)
             continue;
 
         const QmlIR::Object *obj = qmlObjects.at(i);
 
         QQmlPropertyResolver resolver(propertyCache);
-        QQmlPropertyData *defaultProperty = obj->indexOfDefaultPropertyOrAlias != -1 ? propertyCache->parent()->defaultProperty() : propertyCache->defaultProperty();
+        const QQmlPropertyData *defaultProperty = obj->indexOfDefaultPropertyOrAlias != -1 ? propertyCache->parent()->defaultProperty() : propertyCache->defaultProperty();
 
         for (QmlIR::Binding *binding = obj->firstBinding(); binding; binding = binding->next) {
             if (!binding->isValueBinding())
                 continue;
             bool notInRevision = false;
-            QQmlPropertyData *pd = binding->propertyNameIndex != quint32(0) ? resolver.property(stringAt(binding->propertyNameIndex), &notInRevision) : defaultProperty;
+            const QQmlPropertyData *pd = binding->propertyNameIndex != quint32(0) ? resolver.property(stringAt(binding->propertyNameIndex), &notInRevision) : defaultProperty;
             if (pd && pd->isAlias())
                 binding->setFlag(QV4::CompiledData::Binding::IsBindingToAlias);
         }
@@ -737,20 +707,20 @@ void QQmlScriptStringScanner::scan()
 {
     const QMetaType scriptStringMetaType = QMetaType::fromType<QQmlScriptString>();
     for (int i = 0; i < qmlObjects.count(); ++i) {
-        QQmlPropertyCache *propertyCache = propertyCaches->at(i);
+        QQmlPropertyCache::ConstPtr propertyCache = propertyCaches->at(i);
         if (!propertyCache)
             continue;
 
         const QmlIR::Object *obj = qmlObjects.at(i);
 
         QQmlPropertyResolver resolver(propertyCache);
-        QQmlPropertyData *defaultProperty = obj->indexOfDefaultPropertyOrAlias != -1 ? propertyCache->parent()->defaultProperty() : propertyCache->defaultProperty();
+        const QQmlPropertyData *defaultProperty = obj->indexOfDefaultPropertyOrAlias != -1 ? propertyCache->parent()->defaultProperty() : propertyCache->defaultProperty();
 
         for (QmlIR::Binding *binding = obj->firstBinding(); binding; binding = binding->next) {
             if (binding->type() != QV4::CompiledData::Binding::Type_Script)
                 continue;
             bool notInRevision = false;
-            QQmlPropertyData *pd = binding->propertyNameIndex != quint32(0) ? resolver.property(stringAt(binding->propertyNameIndex), &notInRevision) : defaultProperty;
+            const QQmlPropertyData *pd = binding->propertyNameIndex != quint32(0) ? resolver.property(stringAt(binding->propertyNameIndex), &notInRevision) : defaultProperty;
             if (!pd || pd->propType() != scriptStringMetaType)
                 continue;
 
@@ -786,11 +756,12 @@ static bool isUsableComponent(const QMetaObject *metaObject)
     return false;
 }
 
-void QQmlComponentAndAliasResolver::findAndRegisterImplicitComponents(const QmlIR::Object *obj, QQmlPropertyCache *propertyCache)
+void QQmlComponentAndAliasResolver::findAndRegisterImplicitComponents(
+        const QmlIR::Object *obj, const QQmlPropertyCache::ConstPtr &propertyCache)
 {
     QQmlPropertyResolver propertyResolver(propertyCache);
 
-    QQmlPropertyData *defaultProperty = obj->indexOfDefaultPropertyOrAlias != -1 ? propertyCache->parent()->defaultProperty() : propertyCache->defaultProperty();
+    const QQmlPropertyData *defaultProperty = obj->indexOfDefaultPropertyOrAlias != -1 ? propertyCache->parent()->defaultProperty() : propertyCache->defaultProperty();
 
     for (QmlIR::Binding *binding = obj->firstBinding(); binding; binding = binding->next) {
         if (binding->type() != QV4::CompiledData::Binding::Type_Object)
@@ -812,7 +783,7 @@ void QQmlComponentAndAliasResolver::findAndRegisterImplicitComponents(const QmlI
             continue;
         // if here, not a QQmlComponent, so needs wrapping
 
-        QQmlPropertyData *pd = nullptr;
+        const QQmlPropertyData *pd = nullptr;
         if (binding->propertyNameIndex != quint32(0)) {
             bool notInRevision = false;
             pd = propertyResolver.property(stringAt(binding->propertyNameIndex), &notInRevision);
@@ -825,9 +796,9 @@ void QQmlComponentAndAliasResolver::findAndRegisterImplicitComponents(const QmlI
         // If the version is given, use it and look up by QQmlType.
         // Otherwise, make sure we look up by metaobject.
         // TODO: Is this correct?
-        QQmlRefPointer<QQmlPropertyCache> pc = pd->typeVersion().hasMinorVersion()
-                ? enginePrivate->rawPropertyCacheForType(pd->propType(), pd->typeVersion())
-                : enginePrivate->rawPropertyCacheForType(pd->propType());
+        QQmlPropertyCache::ConstPtr pc = pd->typeVersion().hasMinorVersion()
+                ? QQmlMetaType::rawPropertyCacheForType(pd->propType(), pd->typeVersion())
+                : QQmlMetaType::rawPropertyCacheForType(pd->propType());
         const QMetaObject *mo = pc ? pc->firstCppMetaObject() : nullptr;
         while (mo) {
             if (mo == &QQmlComponent::staticMetaObject)
@@ -864,9 +835,9 @@ void QQmlComponentAndAliasResolver::findAndRegisterImplicitComponents(const QmlI
         qmlObjects->append(syntheticComponent);
         const int componentIndex = qmlObjects->count() - 1;
         // Keep property caches symmetric
-        QQmlRefPointer<QQmlPropertyCache> componentCache
-                = enginePrivate->cache(&QQmlComponent::staticMetaObject);
-        propertyCaches.append(componentCache.data());
+        QQmlPropertyCache::ConstPtr componentCache
+                = QQmlMetaType::propertyCache(&QQmlComponent::staticMetaObject);
+        propertyCaches.append(componentCache);
 
         QmlIR::Binding *syntheticBinding = pool->New<QmlIR::Binding>();
         *syntheticBinding = *binding;
@@ -899,8 +870,8 @@ bool QQmlComponentAndAliasResolver::resolve(int root)
         const bool isInlineComponentRoot
                 = obj->flags & QV4::CompiledData::Object::IsInlineComponentRoot;
         const bool isPartOfInlineComponent
-                = obj->flags & QV4::CompiledData::Object::InPartOfInlineComponent;
-        QQmlPropertyCache *cache = propertyCaches.at(i);
+                = obj->flags & QV4::CompiledData::Object::IsPartOfInlineComponent;
+        QQmlPropertyCache::ConstPtr cache = propertyCaches.at(i);
 
         bool isExplicitComponent = false;
         if (obj->inheritedTypeNameIndex) {
@@ -1133,7 +1104,7 @@ QQmlComponentAndAliasResolver::resolveAliasesInObject(int objectIndex,
         if (property.isEmpty()) {
             alias->setFlag(QV4::CompiledData::Alias::AliasPointsToPointerObject);
         } else {
-            QQmlPropertyCache *targetCache = propertyCaches.at(targetObjectIndex);
+            QQmlPropertyCache::ConstPtr targetCache = propertyCaches.at(targetObjectIndex);
             if (!targetCache) {
                 *error = qQmlCompileError(
                         alias->referenceLocation,
@@ -1143,7 +1114,7 @@ QQmlComponentAndAliasResolver::resolveAliasesInObject(int objectIndex,
 
             QQmlPropertyResolver resolver(targetCache);
 
-            QQmlPropertyData *targetProperty = resolver.property(property.toString());
+            const QQmlPropertyData *targetProperty = resolver.property(property.toString());
 
             // If it's an alias that we haven't resolved yet, try again later.
             if (!targetProperty) {
@@ -1191,7 +1162,7 @@ QQmlComponentAndAliasResolver::resolveAliasesInObject(int objectIndex,
                             auto binding = *it;
                             if (compiler->stringAt(binding.propertyNameIndex) == property) {
                                 resolver = QQmlPropertyResolver(propertyCaches.at(binding.value.objectIndex));
-                                QQmlPropertyData *actualProperty = resolver.property(subProperty.toString());
+                                const QQmlPropertyData *actualProperty = resolver.property(subProperty.toString());
                                 if (actualProperty) {
                                     propIdx = QQmlPropertyIndex(propIdx.coreIndex(), actualProperty->coreIndex());
                                     isDeepAlias = true;
@@ -1273,14 +1244,14 @@ bool QQmlDeferredAndCustomParserBindingScanner::scanObject(
         return scanObject(componentBinding->value.objectIndex, ScopeDeferred::False);
     }
 
-    QQmlPropertyCache *propertyCache = propertyCaches->at(objectIndex);
+    QQmlPropertyCache::ConstPtr propertyCache = propertyCaches->at(objectIndex);
     if (!propertyCache)
         return true;
 
     QString defaultPropertyName;
-    QQmlPropertyData *defaultProperty = nullptr;
+    const QQmlPropertyData *defaultProperty = nullptr;
     if (obj->indexOfDefaultPropertyOrAlias != -1) {
-        QQmlPropertyCache *cache = propertyCache->parent().data();
+        const QQmlPropertyCache *cache = propertyCache->parent().data();
         defaultPropertyName = cache->defaultPropertyName();
         defaultProperty = cache->defaultProperty();
     } else {
@@ -1395,24 +1366,8 @@ bool QQmlDeferredAndCustomParserBindingScanner::scanObject(
             }
             isDeferred = true;
         } else if (!deferredPropertyNames.isEmpty() && deferredPropertyNames.contains(name)) {
-            if (seenSubObjectWithId) {
-                qWarning("Binding on %s is not deferred as requested by the DeferredPropertyNames "
-                         "class info because one or more of its sub-objects contain an id.",
-                         qPrintable(name));
-            } else if (binding->type() == Binding::Type_GroupProperty) {
-                // The binding may already be deferred via the surrounding scope.
-                // e.g. PropertyChanges { control.contentItem.opacity: 0.75 }
-                // Here, contentItem is a group property which prevents its deferral. But as
-                // control is already deferred by being a generalized group property, there is
-                // no point in warning here.
-                if (scopeDeferred == ScopeDeferred::False) {
-                    qWarning("Binding on %s is not deferred as requested by the "
-                             "DeferredPropertyNames class info because it constitutes a group "
-                             "property.", qPrintable(name));
-                }
-            } else {
+            if (!seenSubObjectWithId && binding->type() != Binding::Type_GroupProperty)
                 isDeferred = true;
-            }
         }
 
         if (binding->type() >= Binding::Type_Object) {
@@ -1447,7 +1402,7 @@ void QQmlDefaultPropertyMerger::mergeDefaultProperties()
 
 void QQmlDefaultPropertyMerger::mergeDefaultProperties(int objectIndex)
 {
-    QQmlPropertyCache *propertyCache = propertyCaches->at(objectIndex);
+    QQmlPropertyCache::ConstPtr propertyCache = propertyCaches->at(objectIndex);
     if (!propertyCache)
         return;
 
